#
# Selection functions of the tclabc GUI.
#
# This file is part of the tclabc package.
# See the file tkabc.tcl for more information.
#
# Copyright (C) 1999-2005, Jean-Franois Moine.
#
# Contact: mailto:moinejf@free.fr
# Original site: http://moinejf.free.fr/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

proc clip-update {} {
# update the clipboard when something has changed in the selection
    global selection tune
#puts "clip-update"
#    if {[lsearch -exact $selection(syms) [abc go] < 0} return
    # sort the symbols per voice
    set prev_sym {}
    for {set v 0} {$v <= $tune(nvoice)} {incr v} {
	set vsel(v$v) {}
    }
    set vtag {}
    set multi 0
    foreach s $selection(syms) {
	# !!! the voice tag is the 1st tag !!!
	# !!! the symbol address is the 2nd tag !!!
	set tag [.f.c gettags $s]
	if {[string compare $vtag [lindex $tag 0]] != 0 &&
		[string length $vtag] != 0} {
	    set multi 1
	}
	set vtag [lindex $tag 0]
	set sym [lindex $tag 1]
	if {[string compare $prev_sym $sym] != 0} {
	    lappend vsel($vtag) $sym
	    set prev_sym $sym
	}
    }
    # create the clipboard in ABC
    set newsel {}
    set clipboard {}
    for {set v 0} {$v <= $tune(nvoice)} {incr v} {
	if {[llength $vsel(v$v)] == 0} continue
	if {$multi} {
	    if {[llength $newsel] != 0 &&
		[string index $clipboard [expr {[string length $clipboard] - 1}]] != "\n"} {
		append clipboard "\n"
	    }
	    abc go $v
	    set voice [abc voice get]
	    append clipboard "V:[lindex $voice 0]\n"
	}
	append clipboard "L:1/8\n"
	foreach sym $vsel(v$v) {
	    lappend newsel $sym
	    append clipboard [abc dump $sym]
	}
    }
    # memorize the modified list of the selected symbols
    set selection(syms) $newsel
    # copy the selection to the clipboard
    clipboard clear
    clipboard append -format STRING $clipboard
    selection own -command sel-clear .
}

proc mark-exch {} {
# exchange dot and mark
    global selection
    set mark $selection(mark)
    mark-set
    abc go $mark
    cursor-display
}

proc sel-delete {} {
# delete the selection
    global selection tune
    if {[llength $selection(syms)] == 0} return
    foreach s $selection(syms) {
	.f.c delete $s
	abc go $s
	abc delete
    }
    sel-clear
    set s [abc go]
    redisplay 0
    abc go $s
    change-set 1
    cursor-display
}

proc sel-extend {x y} {
# extend the selection - bound to B1-Motion
    global selection voicerect
#    puts stderr "sel-extend $x $y"
    if {[info exists selection(xy)]} {
	set c $selection(xy)
    } else {
	if {$selection(rect) == 0} return
	set c [.f.c coords $selection(rect)]
    }
    foreach [list x1 y1] $c break
    set x2 [.f.c canvasx $x]
    set y2 [.f.c canvasy $y]
    #fixme: have x1 < x2 ?
    if {$selection(rect) == 0} {
	if {$x2 < $x1 + 3 && $y2 > $y1 - 3 && $y2 < $y1 + 3} return
	set selection(rect) [.f.c create rectangle $x1 $y1 $x2 $y2 \
		-outline black -fill DarkSeaGreen2]
	.f.c raise $selection(rect) $voicerect
	set selection(syms) {}
	unset selection(xy)
    } else {
	.f.c coords $selection(rect) $x1 $y1 $x2 $y2
    }
}

proc sel-fetch {offs maxc} {
# return the selection
#puts "sel-fetch"
	return [selection get -selection CLIPBOARD]
}

proc sel-insert {x y} {
# insert the selection after the current symbol or mouse position
    global tune
    if {[catch {selection get} tmp] &&
	[catch {selection get -type TEXT} tmp] &&
	[catch {selection get -selection CLIPBOARD} tmp] &&
	[catch {selection get -selection CLIPBOARD \
		-type STRING} tmp] } return
    if {$x != 0} {
	# paste from mouse: search the nearest symbol
	for {set h 10} {$h < 100} {incr h 10} {
	    set item [.f.c find closest $x $y $h]
	    if {[llength $item] != 0} {
		set tags [.f.c gettags $item]
		if {[llength $tags] > 1} break
	    }
	}
	abc go [lindex $tags 1]	; # symbol is the 2nd tag
	if {[get-xoffset] > $x} {
	    abc go prev
	}
    }
#puts "sel:$tmp"
#fixme: split into voices and dispatch
    set s [abc go]
    abc include $tmp
    abc go $tune(startsym)
    redisplay 1
#fixme: bad cursor position - should be at the end of inserted sequence
    abc go $s
    abc go next
    change-set 1
    cursor-display
}

proc sel-start {x y} {
# start the selection - bound to <Button-1>
    global selection voicerect
#    puts stderr "sel-start $x $y"
    set selection(xy) [list [.f.c canvasx $x] [.f.c canvasy $y]]
#    sel-clear
#    set selection(rect) [.f.c create rectangle $x $y $x $y \
#	-outline black -fill DarkSeaGreen2]
#    .f.c raise $selection(rect) $voicerect
#    focus .f.c
}

proc sel-stop {x y} {
# stop the selection - bound to ButtonRelease-1 and ButtonRelease-3
    global selection tune voicerect
#puts "sel-stop $x $y"
    if {[info exists selection(xy)]} {
	set c $selection(xy)
    } else {
	if {$selection(rect) == 0} return
	set c [.f.c coords $selection(rect)]
    }
    foreach [list x1 y1] $c break
    set x2 [.f.c canvasx $x]
    set y2 [.f.c canvasy $y]
    # get the items which are in the selection
    set selected [.f.c find enclosed $x1 $y1 $x2 $y2]
    if {[llength $selected] == 0} {
	set selected [.f.c find overlapping $x1 $y1 $x2 $y2]
	# remove the background elements from the selection
	# (they have less than 2 tags)
	set newsel {}
	foreach s $selected {
	    set t [.f.c gettags $s]
	    if {[llength $t] > 1} {
		lappend newsel $s
	    }
	}
	set selected $newsel
	if {[llength $selected] == 0} {
	    set selected [.f.c find closest $x2 $y2 10]
	    if {[llength $selected] != 0} {
		set t [.f.c gettags $selected]
		if {[llength $t] <= 1} {
		    set selected {}
		}
	    }
	}
    }
    # if no symbol or simple pointing (no mouse move), done
    if {[llength $selected] == 1 &&
	    $x2 < $x1 + 3 && $y2 > $y1 - 3 && $y2 < $y1 + 3} {
	abc go [lindex [.f.c gettags $selected] 1]
	set selected {}
    }
    if {[llength $selected] == 0} {
	# no element found, stay where we are
	if {[llength $selection(syms)] == 0 && $selection(rect) != 0} {
	    .f.c delete $selection(rect)
	    set selection(rect) 0
	}
	catch {unset selection(xy)}
	sym-sel
	return
    }
    set selection(syms) $selected
    clip-update
    # select the first symbol
    abc go [lindex $selection(syms) 0]
    sym-sel
    # update the selection rectangle
    foreach {x1 y1 x2 y2} [eval .f.c bbox $selection(syms)] break
    incr x1 -5
    incr y1 -5
    incr x2 4
    incr y2 4
    if {$selection(rect) == 0} {
	set selection(rect) [.f.c create rectangle $x1 $y1 $x2 $y2 \
		-outline black -fill DarkSeaGreen2]
	.f.c raise $selection(rect) $voicerect
	unset selection(xy)
    } else {
	.f.c coords $selection(rect) $x1 $y1 $x2 $y2
    }
}
